"""
A package for working with datasets in this benchmark: loading datasets as PyTorch
Datasets, numpy arrays, or pandas dataframes; evaluating models using a unified set
of metrics
"""

from . import datasets
